#! /usr/bin/env bash

# Copyright (c) 2018-2020 NVIDIA Corporation
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
# Released under the Apache License v2.0 with LLVM Exceptions.
# See https://llvm.org/LICENSE.txt for license information.

################################################################################
# Thrust and CUB build script for gpuCI (CPU-only)
################################################################################

export PARALLEL_LEVEL=${PARALLEL_LEVEL:-4}

source ${WORKSPACE}/ci/common/build.bash

